/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
#ifdef PASTIX_DYNSCHED
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <assert.h>
#include <sys/stat.h>

#include "common_pastix.h"
#include "queue.h"
#include "extendVector.h"
#include "cand.h"
#include "ftgt.h"
#include "symbol.h"
#include "simu.h"
#include "cost.h"
#include "elimin.h"
#include "bulles.h"
#include "param_blend.h"
#include "blendctrl.h"
#include "dof.h"
#include "csc.h"
#include "updown.h"
#include "solver.h"
#include "costfunc.h"
#include "splitpartlocal.h"

/* #define DEBUG_BUBBLESPLIT */
#define ONE_COST
/* #define TEST_REPART */
#define CROSS_TOLERANCE 0.5

extern void setTreeLevel    (Cand *, const EliminTree *);
extern void setTreeCostLevel(Cand *, const EliminTree *, const CostMatrix *);

int VerifTaskRepartition(BlendCtrl *ctrl, SimuCtrl * simuctrl);

void splitPartLocal(BlendCtrl *ctrl, SimuCtrl * simuctrl, SymbolMatrix *symbmtx, const Dof *dofptr)
{

  double *bubble_cost = NULL;
  INT     i, j, size;
    
  /* On met a zero le cot de toutes les oprations qui ne se font pas sur le processus MPI courant */
  subtreeUpdateCostLocal(ROOT(ctrl->etree), ctrl, symbmtx, simuctrl, dofptr, ctrl->clustnum);

  setTreeLevel(ctrl->candtab, ctrl->etree);
  if(ctrl->option->costlevel)
    setTreeCostLevel(ctrl->candtab, ctrl->etree, ctrl->costmtx);

  /* Initialisation de la structure de d'arbre de bulles */
  Bubble_InitTree(ctrl->btree, ctrl->thrdlocnbr);

  MALLOC_INTERN(bubble_cost, ctrl->btree->nodemax, double);
  for (i=0; i< ctrl->btree->nodemax; i++)
    bubble_cost[i] = 0.0;

  /* Repartition symbolic matrix using proportionnal mapping method */
  propMappTreeLocal(ctrl, symbmtx, simuctrl, bubble_cost);

  /* construction de l'arbre de bulles */
  Bubble_BuildTree(ctrl->btree);

  if (ctrl->option->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
    {
      FILE *out;
      char  filename[18];
      sprintf(filename,"BubbleTree.%02d.dot", (int)ctrl->clustnum);
      filename[17] = '\0';
      OUT_OPENFILEINDIR(ctrl->option->iparm, out, filename, "w");
      Bubble_Print(ctrl->btree, bubble_cost, out);
      OUT_CLOSEFILEINDIR(out);
    }

  ctrl->bublnbr = ctrl->btree->nodenbr;

  /*     VerifTaskRepartition(ctrl, simuctrl); */
    
  simuctrl->clustab[ctrl->clustnum].fprocnum = 0;
  simuRealloc(simuctrl, ctrl->procnbr, ctrl->bublnbr);

#ifdef TEST_REPART
  for (i=0; i<ctrl->btree->leavesnbr; i++)
    {
      INT    task;
      double key;
      j = BFATHER(ctrl->btree, i);
      if (j == -1) continue;

      while(queueSize(ctrl->btree->nodetab[i].taskheap))
	{
	  task = queueGet2(ctrl->btree->nodetab[i].taskheap, &key, NULL);
	  queueAdd(ctrl->btree->nodetab[j].taskheap, task, key);
	}
    }    
#endif

  for (i=0; i<ctrl->bublnbr; i++)
    {
      size = queueSize(ctrl->btree->nodetab[i].taskheap);

      /* fprintf(stdout,"La bulle %3d doit recevoir %4d taches en %g\n", i, size, bubble_cost[i]);*/
      extendint_Init(simuctrl->proctab[i].tasktab, size);
	
      for(j=0; j<size; j++)
	{
	  extendint_Add(simuctrl->proctab[i].tasktab, queueGet(ctrl->btree->nodetab[i].taskheap));
	}
    }
    
  memFree_null(bubble_cost);
}


/*+ Repartition symbolic matrix using proportionnal mapping method +*/
void propMappTreeLocal(BlendCtrl *ctrl, const SymbolMatrix *symbmtx, const SimuCtrl * simuctrl, double * bubble_cost)
{
  double isocost;
  INT pr;
  double *cost_remain = NULL; 
  MALLOC_INTERN(cost_remain, (ctrl->thrdlocnbr), double);
  isocost     = ctrl->costmtx->cblktab[ROOT(ctrl->etree)].subtree / ctrl->thrdlocnbr;
  
  /* compute cost to get in the elimination tree for each processor */
  for(pr=0;pr<ctrl->thrdlocnbr;pr++)
    cost_remain[pr] = isocost;

  propMappSubtreeLocalNC(ctrl, symbmtx, simuctrl, ROOT(ctrl->etree), 0, ctrl->thrdlocnbr-1, cost_remain, bubble_cost);

#ifdef ONE_COST
  memFree_null(cost_remain);
#endif
}


void propMappSubtreeLocalNC(BlendCtrl *ctrl, const SymbolMatrix *symbmtx, const SimuCtrl * simuctrl, INT rootnum, 
			    INT fcandnum, INT lcandnum, double *cost_remain, double *bubble_cost)
{
  INT i, j;
  INT candnbr;
  INT fcand = 0;
  INT lcand = 0;
  INT sonsnbr;
  INT bubblenum;
  INT tasknum;
  double aspt_cost;
  double cumul_cost;
  double epsilon;
  double soncost;
  Queue *queue_tree;
  Queue *queue_proc;
#ifndef ONE_COST
  INT p;
  double isocost;
  double *sub_cost_remain;
#endif

  candnbr = lcandnum - fcandnum + 1;

  print_debug(DBG_BUBBLESPLIT,"Task1 %4d : (%f)\n", (int)rootnum, ctrl->costmtx->cblktab[rootnum].total);
  for(i=fcandnum; i<lcandnum+1; i++)
    print_debug(DBG_BUBBLESPLIT,"            (%d, %f)\n", (int)i, cost_remain[i-fcandnum]);

  /* Si le sous-arbre est trait par un autre processus MPI */
  if (ctrl->costmtx->cblktab[rootnum].subtree == 0.0)
    {
#ifndef ONE_COST      
      memFree(cost_remain);
#endif
      print_debug(DBG_BUBBLESPLIT,"            (Trait par un autre proc)\n");

      return;
    }
    
  
  /* Creation de la nouvelle bulle si besoin */
  bubblenum = Bubble_Add(ctrl->btree, fcandnum, lcandnum, 
			 ctrl->candtab[rootnum].costlevel, 
			 ctrl->candtab[rootnum].treelevel);
  
  /* On ajoute le noeud courant  la bulle adequate */
  if (ctrl->costmtx->cblktab[rootnum].total != 0.0 )
    {
      tasknum = simuctrl->bloktab[symbmtx->cblktab[rootnum].bloknum].tasknum;
      ASSERTDBG(tasknum != -1, MOD_BLEND);

      if (ctrl->candtab[rootnum].distrib == D1)
        {
	  ASSERTDBG(simuctrl->tasktab[tasknum].taskid == COMP_1D, MOD_BLEND);
          queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, tasknum, simuctrl->tasktab[tasknum].prionum);
        }
      /* D2 */
      else 
        {
          /* Diag */
	  if (ctrl->proc2clust[simuctrl->blprtab[symbmtx->cblktab[rootnum].bloknum]] == ctrl->clustnum)
	    {
	      tasknum = simuctrl->bloktab[symbmtx->cblktab[rootnum].bloknum].tasknum;
	      ASSERTDBG(tasknum != -1, MOD_BLEND);
	      ASSERTDBG(simuctrl->tasktab[tasknum].taskid == DIAG, MOD_BLEND);
	      ASSERTDBG(ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] == ctrl->clustnum, MOD_BLEND);

	      queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, 
		       tasknum, simuctrl->tasktab[tasknum].prionum);
	    }

          /* E1 tasks */
          for(i=symbmtx->cblktab[rootnum].bloknum+1;i<symbmtx->cblktab[rootnum+1].bloknum;i++)
            {
              tasknum = simuctrl->bloktab[i].tasknum;
	      if (ctrl->proc2clust[simuctrl->blprtab[i]] == ctrl->clustnum)
		{
		  ASSERTDBG(tasknum != -1, MOD_BLEND);
		  ASSERTDBG(simuctrl->tasktab[tasknum].taskid == E1, MOD_BLEND);
		  ASSERTDBG(ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] == ctrl->clustnum, MOD_BLEND);
		  
		  queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, tasknum, 
			   simuctrl->tasktab[tasknum].prionum);
		}
	      else
		{
		  ASSERTDBG(ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] != ctrl->clustnum, MOD_BLEND);
		}


              /* E2 tasks */
              for(j=i;j<symbmtx->cblktab[rootnum+1].bloknum;j++)
                {
                  tasknum++;
		  if (ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] == ctrl->clustnum)
		    queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, tasknum, 
			     simuctrl->tasktab[tasknum].prionum);

		  ASSERTDBG(simuctrl->tasktab[tasknum].taskid == E2, MOD_BLEND);
                }
            }
        }

      ctrl->candtab[rootnum].cluster = bubblenum;
      bubble_cost[bubblenum] += ctrl->costmtx->cblktab[rootnum].total;
      
      if (fcandnum == lcandnum)
	cost_remain[0] -= ctrl->costmtx->cblktab[rootnum].total;
#ifndef ONE_COST
      else
	{
	  /* work that each processor is intended to get from this treenode */
	  isocost = ctrl->costmtx->cblktab[rootnum].total / candnbr;
	  for(i=0;i<candnbr;i++)
	    cost_remain[i] -= isocost;
	}
#endif
    }

  /* this treenode is a leave return */
  if(ctrl->etree->nodetab[rootnum].sonsnbr == 0)
    {
#ifndef ONE_COST
      memFree(cost_remain);
#endif

      print_debug(DBG_BUBBLESPLIT,"            (Noeud feuille)\n");

      return;
    }
    
  /* Cost remaining in the descendance of the treenode after split*/
  aspt_cost = ctrl->costmtx->cblktab[rootnum].subtree - ctrl->costmtx->cblktab[rootnum].total;

  if(cost_remain[0] <= 0)
    fcand = 1;
  else
    fcand = 0;
/*   if(cost_remain[candnbr-1] <= 0) */
/*     candnbr--; */


  /** ASSERT that the sum of cost_remain in used proc equals to
    after split cost **/
  cumul_cost = 0;
  for(i=fcand;i<candnbr;i++)
    cumul_cost += cost_remain[i];
  for(i=fcand;i<candnbr;i++)
    cost_remain[i] *= aspt_cost/cumul_cost;

  /** Compute the minimun participation rate of a candidat processor **/
  epsilon = CROSS_TOLERANCE*aspt_cost/candnbr ;
  
#ifdef DEBUG_BUBBLESPLIT2
  {
    fprintf(stdout, "Rootnum %ld fproc %ld lproc %ld [ ", (long)rootnum, (long)fcandnum, (long)lcandnum);
    for(p=0;p<candnbr;p++)
      fprintf(stdout, "%g ", cost_remain[p]);
    fprintf(stdout, " ]\n");
    fprintf(stdout, " Sons ");
    for(i=0;i<ctrl->etree->nodetab[rootnum].sonsnbr;i++)
      fprintf(stdout, " [%ld, %g] ", i, ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree);
    fprintf(stdout, "\n");
  }
#endif

  lcand = fcand;
  
  /* compute the cand group for each proc */
  sonsnbr = ctrl->etree->nodetab[rootnum].sonsnbr;

  /* allocate queue of proc and subtree */
  MALLOC_INTERN(queue_tree, 1, Queue);
  MALLOC_INTERN(queue_proc, 1, Queue);
  queueInit(queue_tree, sonsnbr);
  queueInit(queue_proc, candnbr);


  /* Fill queue subtree order by cost descending */

  print_debug(DBG_BUBBLESPLIT,"-----------------------------\n");
  print_debug(DBG_BUBBLESPLIT,"Task2 %4d : \n", (int)rootnum);

  for(i=0;i<sonsnbr;i++)
    {
      
      print_debug(DBG_BUBBLESPLIT,"            (%4d, %f)\n",  (int)TSON(ctrl->etree, (int)rootnum, i),
		  ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree);


      /* Le fils est entierement trait sur un autre noeud => on le passe */
      if (ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree == 0.0 )
	continue;
      
      /** Cost in the current subtree to be mapped **/
      cumul_cost = -ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree;
      
      /* Cost of the root node in the subtree */
      soncost    = -ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].total;

#ifdef PASTIX_DYNSCHED_OVERHEAD
      if (-cumul_cost < epsilon)
	propMappSubtreeLocalOn1P(ctrl, symbmtx, simuctrl, TSON(ctrl->etree, rootnum, i), 
				 fcandnum, lcandnum, bubblenum, bubble_cost);
      else
#endif
	queueAdd2(queue_tree, i, cumul_cost, soncost);

    }

  print_debug(DBG_BUBBLESPLIT,"-----------------------------\n");



  /* Proportionnal mapping of the subtree on cand */
  /* Premiere repartition des sous-arbres de cout superieur a un proc */
  while (queueSize(queue_tree) > 0)
    {  
      i = queueGet(queue_tree);

      /** Cost in the current subtree to be mapped **/
      cumul_cost = ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree;

      /* Le sous-arbre peut-etre entierement excut par le premier candidat    */
      /* et par consequence toutes les taches suivantes ne serront plus splitt */
      if (cumul_cost < cost_remain[fcand]){

	cost_remain[fcand] -= cumul_cost;
	bubblenum           = fcandnum + fcand;

	propMappSubtreeLocalOn1P(ctrl, symbmtx, simuctrl, TSON(ctrl->etree, rootnum, i), 
				 bubblenum, bubblenum, bubblenum, bubble_cost);
	bubble_cost[bubblenum] += cumul_cost;

	print_debug(DBG_BUBBLESPLIT,"            (%4d, %f, %d, %d)\n",  (int)TSON(ctrl->etree, rootnum, i), 
		    cumul_cost, 1, (int)bubblenum);

	break;
      }

      /* Update subtree cost and remain cost of fcand proc */
      cumul_cost -= cost_remain[fcand];

#ifdef DEBUG_BUBBLESPLIT
      if(cost_remain[fcand]<=0)
	fprintf(stderr, "candnbr %ld fcand %ld cost_remain[fcand] %g epsilon %g \n", 
		(long)candnbr, (long)fcand, cost_remain[fcand], epsilon);
#endif

      lcand = fcand;
/*       while ((cumul_cost > ((lcand - fcand +1)*epsilon)) &&  */
/* 	     (lcand < candnbr - 1)) */
      while ((cumul_cost > epsilon) && (lcand < candnbr - 1))
	{
	  lcand++;
	  cumul_cost -= cost_remain[lcand];
	  ASSERTDBG(cost_remain[lcand]>0,MOD_BLEND);
	}

#ifndef ONE_COST
      MALLOC_INTERN(sub_cost_remain, lcand-fcand+1, double);
      if (cumul_cost > 0)
	{
	  for(p=0;p<lcand-fcand+1;p++)
	    {
	      sub_cost_remain[p]   = cost_remain[fcand+p] + (cumul_cost / (lcand - fcand + 1));
	      cost_remain[fcand+p] = - cumul_cost / (lcand - fcand + 1);
	    }
	}
      else
	{
	  /* compute the cost to get for each proc cand for this subtree */
	  for(p=0; p<lcand-fcand; p++)
	    {
	      sub_cost_remain[p] = cost_remain[fcand+p];
	      cost_remain[fcand+p] = 0.0;
	    }
	  sub_cost_remain[lcand-fcand] = cost_remain[lcand] + cumul_cost; /* cumul_cost <= 0 */
	  cost_remain[fcand+p]         = - cumul_cost ;
	}
#endif

      /* go on to subtree */
      print_debug(DBG_BUBBLESPLIT,"            (%4d, %f, %d, %d)\n",  (int)TSON(ctrl->etree, rootnum, i), 
		  ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree, (int)(lcand-fcand+1), (int)(fcandnum+fcand));


#ifdef ONE_COST
      propMappSubtreeLocalNC(ctrl, symbmtx, simuctrl, TSON(ctrl->etree, rootnum, i), 
			     fcandnum+fcand, fcandnum+lcand, cost_remain+fcand, bubble_cost);
#else
      propMappSubtreeLocalNC(ctrl, symbmtx, simuctrl, TSON(ctrl->etree, rootnum, i), 
			     fcandnum+fcand, fcandnum+lcand, sub_cost_remain, bubble_cost);
#endif

      print_debug(DBG_BUBBLESPLIT,"Task3 %4d : \n", (int)rootnum);

      
      if (lcand < candnbr - 1)
	fcand = lcand+1;
      else
	break;
    }

  /* Fill queue proc order by remain cost descending */
  for (i=0; i<candnbr; i++)
#ifdef ONE_COST
    queueAdd(queue_proc, i, bubble_cost[i]);
#else
    queueAdd(queue_proc, i, -cost_remain[i]); 
#endif

  /* Affect all subtree executed by only one thread */
  while (queueSize(queue_tree) > 0)
    {  
      i     = queueGet(queue_tree);
      fcand = queueGet(queue_proc);
      
      /** Cost in the current subtree to be mapped **/
      cumul_cost = ctrl->costmtx->cblktab[TSON(ctrl->etree, rootnum, i)].subtree;

      bubblenum = fcandnum + fcand;
      propMappSubtreeLocalOn1P(ctrl, symbmtx, simuctrl, TSON(ctrl->etree, rootnum, i),
			       bubblenum, bubblenum, bubblenum, bubble_cost);

      print_debug(DBG_BUBBLESPLIT,"            (%4d, %f, %d, %d)\n",  (int)TSON(ctrl->etree, rootnum, i), 
		  cumul_cost, 1, (int)bubblenum);

      bubble_cost[bubblenum] += cumul_cost;
      cost_remain[fcand]     -= cumul_cost;

#ifdef ONE_COST
      queueAdd(queue_proc, fcand, bubble_cost[fcand]);
#else
      queueAdd(queue_proc, fcand, -cost_remain[fcand]); 
#endif
    }

  queueExit(queue_tree);
  queueExit(queue_proc);
  memFree(queue_tree);
  memFree(queue_proc);
#ifndef ONE_COST
  memFree(cost_remain);
#endif
  return;
}


void propMappSubtreeLocalOn1P(BlendCtrl *ctrl, const SymbolMatrix *symbmtx, const SimuCtrl *simuctrl,
			      INT rootnum, INT fcandnum, INT lcandnum, INT bubblenum,
			      double *bubble_cost)
{
  INT i, j;
  INT sonsnbr;
  INT tasknum;

  /* Si le sous-arbre est trait par un autre processus MPI */
  if (ctrl->costmtx->cblktab[rootnum].subtree == 0.0)
    return;
  
  /* work that each processor is intended to get from this treenode */
  if (ctrl->costmtx->cblktab[rootnum].total != 0.0 )
    {
      tasknum = simuctrl->bloktab[symbmtx->cblktab[rootnum].bloknum].tasknum;
      ASSERTDBG(tasknum != -1, MOD_BLEND);
      
      if (ctrl->candtab[rootnum].distrib == D1)
        {
	  ASSERTDBG(simuctrl->tasktab[tasknum].taskid == COMP_1D, MOD_BLEND);
          queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, tasknum, simuctrl->tasktab[tasknum].prionum);
        }
      /* D2 */
      else 
        {
          /* Diag */
	  if (ctrl->proc2clust[simuctrl->blprtab[symbmtx->cblktab[rootnum].bloknum]] == ctrl->clustnum)
	    {
	      tasknum = simuctrl->bloktab[symbmtx->cblktab[rootnum].bloknum].tasknum;
	      ASSERTDBG(tasknum != -1, MOD_BLEND);
	      ASSERTDBG(simuctrl->tasktab[tasknum].taskid == DIAG, MOD_BLEND);
	      ASSERTDBG(ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] == ctrl->clustnum, MOD_BLEND);
	      
	      queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, 
		       tasknum, simuctrl->tasktab[tasknum].prionum);
	    }

          /* E1 tasks */
          for(i=symbmtx->cblktab[rootnum].bloknum+1;i<symbmtx->cblktab[rootnum+1].bloknum;i++)
            {
              tasknum = simuctrl->bloktab[i].tasknum;
	      if (ctrl->proc2clust[simuctrl->blprtab[i]] == ctrl->clustnum)
		{
		  ASSERTDBG(tasknum != -1, MOD_BLEND);
		  ASSERTDBG(simuctrl->tasktab[tasknum].taskid == E1, MOD_BLEND);
		  ASSERTDBG(ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] == ctrl->clustnum, MOD_BLEND);
		  
		  queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, tasknum, 
			   simuctrl->tasktab[tasknum].prionum);
		}
	      else
		{
		  ASSERTDBG(ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] != ctrl->clustnum, MOD_BLEND);
		}
	      

              /* E2 tasks */
              for(j=i;j<symbmtx->cblktab[rootnum+1].bloknum;j++)
                {
                  tasknum++;
		  if (ctrl->proc2clust[simuctrl->blprtab[simuctrl->tasktab[tasknum].bloknum]] == ctrl->clustnum)
		    queueAdd(ctrl->btree->nodetab[bubblenum].taskheap, tasknum, 
			     simuctrl->tasktab[tasknum].prionum);
		  
		  ASSERTDBG(simuctrl->tasktab[tasknum].taskid == E2, MOD_BLEND);
                }
            }
        }

      ctrl->candtab[rootnum].cluster = bubblenum;
    }

  sonsnbr = ctrl->etree->nodetab[rootnum].sonsnbr;
  
  /* Fill queue subtree order by cost descending */
  for(i=0;i<sonsnbr;i++)
    propMappSubtreeLocalOn1P(ctrl, symbmtx, simuctrl, TSON(ctrl->etree, rootnum, i), 
			     fcandnum, lcandnum, bubblenum, bubble_cost);
  
  return;
}


int VerifTaskRepartition(BlendCtrl *ctrl, SimuCtrl * simuctrl){

  int i, j, clust;

  for (clust=0; clust<ctrl->clustnbr; clust++){
      if (ctrl->clustnum == clust)
	{
	  int *indice = NULL;
	  int *value  = NULL;
	  int *fin    = NULL;
	  int nbfin=0, nberror;
	  Queue ** copyqueue = NULL;
	  
	  MALLOC_INTERN(indice,    ctrl->proclocnbr, int);
	  MALLOC_INTERN(fin,       ctrl->thrdlocnbr, int);
	  MALLOC_INTERN(value,     ctrl->thrdlocnbr, int);
	  MALLOC_INTERN(copyqueue, ctrl->thrdlocnbr, Queue*);

	  for (i=0; i<ctrl->procnbr; i++)
	    indice[i] = 0;

	  for (i=0; i<ctrl->thrdlocnbr; i++){
	    fin[i] = 1;
	    value[i] = -1;
	    MALLOC_INTERN(copyqueue[i], 1, Queue);
	    copyqueue[i] = queueCopy(copyqueue[i], ctrl->btree->nodetab[i].taskheap);
	  }

	  while (nbfin < ctrl->thrdlocnbr ){

	    for (i=0; i<ctrl->thrdlocnbr ; i++){
	      if (fin[i] && value[i] < 0)
		{
		  if (queueSize(copyqueue[i]) > 0)
		    value[i] = queueGet(copyqueue[i]);
		  else{
		    value[i] = -1;
		    fin[i]   = 0;
		    nbfin++;
		  }
		}
	    }
	
	    if (!(nbfin<ctrl->thrdlocnbr))
	      break;

	    printf("tasktab ");
	    for (i=simuctrl->clustab[ctrl->clustnum].fprocnum; 
		 i<simuctrl->clustab[ctrl->clustnum].lprocnum+1; i++)
	      printf("%d ", (indice[i]<extendint_Size(simuctrl->proctab[i].tasktab))?
		     extendint_Read(simuctrl->proctab[i].tasktab, indice[i]):-1);
	    printf("\n");
	
	    printf("queue ");
	    for (i=0; i<ctrl->thrdlocnbr; i++)
	      printf("%d ", value[i]);
	    printf("\n");
	
	    nberror = 0;
	    for(i=simuctrl->clustab[ctrl->clustnum].fprocnum; 
		i<simuctrl->clustab[ctrl->clustnum].lprocnum+1; i++)
	      {
		for(j=0; j<ctrl->thrdlocnbr; j++)
		  {
		    if (value[j] != -1 &&
			indice[i] < extendint_Size(simuctrl->proctab[i].tasktab) &&
			value[j] == extendint_Read(simuctrl->proctab[i].tasktab, indice[i]))
		      {
			indice[i]++;
			value[j] = -2;
			break;
		      }
		    else
		      nberror++;
		  }
	      }
	
	    if (!(nberror < ctrl->proclocnbr*ctrl->thrdlocnbr)){
	      printf("tasktab ");
	      for(i=simuctrl->clustab[ctrl->clustnum].fprocnum; 
		  i<simuctrl->clustab[ctrl->clustnum].lprocnum+1; i++)
		printf("%d ", extendint_Read(simuctrl->proctab[i].tasktab, indice[i]));
	      printf("\n");
	    
	      printf("queue ");
	      for (i=0; i<ctrl->thrdlocnbr; i++)
		printf("%d ", value[i]);
	      printf("\n");

	    }
	    ASSERT(nberror < ctrl->procnbr*ctrl->thrdlocnbr, MOD_BLEND);
	    
	  }
	}
      else
	sleep(10);
    }
    return 1;
}

#endif
